#include "DXUT.h"
#include "DxaImport.h"
#include <assimp.hpp>      // C++ importer interface
#include <aiScene.h>       // Output data structure
#include <aiPostProcess.h> // Post processing flags

Dxa::Import::Import(ID3D11Device* device)
	:Dxa::Sas(device)
{
	#pragma region 6.4
	camera.SetViewParams( &D3DXVECTOR3(0, 3, -10), &D3DXVECTOR3(0, 3, 0) );
	#pragma endregion set up camera view
}

void Dxa::Import::loadEffect()
{
	#pragma region 6.5
	ID3DBlob* compiledEffect = NULL;
	ID3DBlob* compilationErrors = NULL;
	HRESULT hr = D3DX11CompileFromFileW(
			L"fx/envmapped.fx", NULL, NULL, NULL,
			"fx_5_0", 0, 0, NULL, &compiledEffect, &compilationErrors, NULL);
	if(hr != S_OK)
	{
		if(compilationErrors != NULL)
			MessageBoxA( NULL, 
				(LPSTR)compilationErrors->GetBufferPointer(),
				"Failed to create effect from file!", MB_OK);
		else
			MessageBoxA( NULL, 
				"File cound not be opened",
				"Failed to create effect from file!", MB_OK);
		exit(-1);
	}
	#pragma endregion compile effect file with error handling
	#pragma region 6.6
	hr = D3DX11CreateEffectFromMemory(
		compiledEffect->GetBufferPointer(), 
		compiledEffect->GetBufferSize(),
		0, device, &effect);
	if(hr != S_OK)
	{
		MessageBoxA( NULL, 
			"CreateEffectFromMemory failed",
			"Failed to create effect from file!", MB_OK);
		exit(-1);
	}
	#pragma endregion create effect with error handling
}

HRESULT Dxa::Import::createResources()
{
	Dxa::Sas::createResources();

	addButton(L"Import asset", 5, 50, 125, 22, 0, false, new OpenFileFunctor(this));
	addButton(L"Save geometry", 5, 80, 125, 22, 0, false, new SaveFileFunctor(this));

	inputLayoutManager = new InputLayoutManager();

	#pragma region 6.7
	mesh = NULL;
	//Mesh::createMeshFromFile("media/bigbunny.dgb", device, &mesh);
		Assimp::Importer importer;

		const aiScene* assScene = importer.ReadFile( "media/polyhedrons_ts.3ds",
			aiProcess_FlipWindingOrder		|
			aiProcess_CalcTangentSpace		| 
			aiProcess_Triangulate			|
			aiProcess_JoinIdenticalVertices	|
			aiProcess_SortByPType);
  
		// if the import failed
		if( !assScene || !assScene->HasMeshes() || assScene->mNumMeshes == 0) 
		{
			mesh = NULL;
			return E_FAIL;
		}

		aiMesh* assMesh = assScene->mMeshes[0];

		Mesh::importMesh(assMesh, device, &mesh, &vertexData, &indexData);
	if(mesh == NULL)
	{
		MessageBoxA( NULL, 
			"Mesh could not be loaded.",
			"Failed to create mesh from file!", MB_OK);
		exit(-1);
	}
	#pragma endregion create mesh from dgb file

	#pragma region 8.5
	D3DX11_PASS_DESC basicPassDesc;
	effect->GetTechniqueByName("envmapped")->GetPassByName("envmapped")->GetDesc(&basicPassDesc);
	//mesh->createInputLayout(device, &basicPassDesc, &inputLayout);
	inputLayout = inputLayoutManager->getCompatibleInputLayout(device, basicPassDesc, mesh, false);
	#pragma endregion create input layout for rendering the mesh with the textured technique

	#pragma region 8.3
	D3DX11CreateShaderResourceViewFromFile( device, L"Media/uvGrid.jpg", NULL, NULL, &kdTextureSrv, NULL);
	#pragma endregion load texture, create shader resource view

	#pragma region 10.1
	D3DX11CreateShaderResourceViewFromFile( device, L"Media/cloudyNoon.dds", NULL, NULL, &envTextureSrv, NULL);
	#pragma endregion load environment cube texture, create shader resource view

	#pragma region 12.4
	Mesh::createQuad(device, &backgroundQuad);
	D3DX11_PASS_DESC backgroundPassDesc;
	effect->GetTechniqueByName("envmapped")->GetPassByName("background")->GetDesc(&backgroundPassDesc);
	//backgroundQuad->createInputLayout(device, &backgroundPassDesc, &backgroundInputLayout);
	backgroundInputLayout = inputLayoutManager->getCompatibleInputLayout(device, backgroundPassDesc, backgroundQuad, false);

	#pragma endregion background quad mesh and input layout for the background technique pass

	return S_OK;
}

HRESULT Dxa::Import::releaseResources()
{
	#pragma region 12.5
	delete backgroundQuad;
	#pragma endregion release background quad mesh and input layout

	#pragma region 10.1
	envTextureSrv->Release();
	#pragma endregion release environment cube texture srv
	#pragma region 8.4
	kdTextureSrv->Release();
	#pragma endregion release texture srv
	#pragma region 6.9
	delete mesh;
	#pragma endregion release resources

	delete inputLayoutManager;

	return Dxa::Sas::releaseResources();
}

HRESULT Dxa::Import::createSwapChainResources()
{
	Dxa::Sas::createSwapChainResources();

	#pragma region 6.10
	camera.SetProjParams(1.58, (float)backbufferSurfaceDesc.Width / backbufferSurfaceDesc.Height, 0.1, 1000 );
	#pragma endregion set camera aspect ratio
	return S_OK;
}

HRESULT Dxa::Import::releaseSwapChainResources()
{
	#pragma region 6.11
	#pragma endregion do nothing
	
	return Dxa::Sas::releaseSwapChainResources();
}

#pragma region 6.12
void Dxa::Import::animate(double dt, double t)
{
	Dxa::Sas::animate(dt, t);
	camera.FrameMove(dt);
}

bool Dxa::Import::processMessage( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	if(Dxa::Sas::processMessage(hWnd, uMsg, wParam, lParam))
		return true;
	return camera.HandleMessages(hWnd, uMsg, wParam, lParam);
}
#pragma endregion pass messages to camera object

void Dxa::Import::render(ID3D11DeviceContext* context)
{
	#pragma region 6.13
	float clearColor[4] = { 0.3f, 0.2f, 0.9f, 0.0f };

	ID3D11RenderTargetView* defaultRtv = DXUTGetD3D11RenderTargetView();
	ID3D11DepthStencilView* defaultDsv = DXUTGetD3D11DepthStencilView();
	context->ClearRenderTargetView( defaultRtv, clearColor );
	context->ClearDepthStencilView( defaultDsv, D3D11_CLEAR_DEPTH, 1.0, 0 );
	#pragma endregion clear backbuffer and depth

	#pragma region 6.14
	D3DXMATRIX modelMatrix, modelMatrixInverse;
	D3DXMatrixIdentity(&modelMatrix);
	D3DXMatrixInverse(&modelMatrixInverse, NULL, &modelMatrix);
	D3DXMATRIX viewMatrix = *camera.GetViewMatrix();
	D3DXMATRIX projMatrix = *camera.GetProjMatrix();
	D3DXMATRIX modelViewProjMatrix = modelMatrix * viewMatrix * projMatrix;
	#pragma endregion compute matrices

	#pragma region 6.15
	effect->GetVariableByName("modelMatrix")->AsMatrix()->SetMatrix((float*)&modelMatrix);
	effect->GetVariableByName("modelMatrixInverse")->AsMatrix()->SetMatrix((float*)&modelMatrixInverse);
	effect->GetVariableByName("modelViewProjMatrix")->AsMatrix()->SetMatrix((float*)&modelViewProjMatrix);
	#pragma endregion set global effect variables

	#pragma region 12.6
	D3DXMATRIX eyeMatrix;
	D3DXMatrixTranslation(&eyeMatrix, camera.GetEyePt()->x, camera.GetEyePt()->y, camera.GetEyePt()->z);
	D3DXMATRIX eyeViewProjMatrix = eyeMatrix * viewMatrix * projMatrix;
	D3DXMATRIX eyeViewProjMatrixInverse;
	D3DXMatrixInverse(&eyeViewProjMatrixInverse, NULL, &eyeViewProjMatrix);
	effect->GetVariableByName("eyeViewProjMatrixInverse")->AsMatrix()->SetMatrix((float*)&eyeViewProjMatrixInverse);
	#pragma endregion compute screen-to-viewdir matrix

	#pragma region 10.2
	effect->GetVariableByName("eyePos")->AsVector()->SetFloatVector((float*)camera.GetEyePt());
	effect->GetVariableByName("kdTexture")->AsShaderResource()->SetResource(kdTextureSrv);
	effect->GetVariableByName("envTexture")->AsShaderResource()->SetResource(envTextureSrv);
	context->IASetInputLayout(inputLayout);
	effect->GetTechniqueByName("envmapped")->GetPassByName("envmapped")->Apply(0, context);
	mesh->draw(context);
	#pragma endregion draw mesh using technique basic

	#pragma region 12.7
	context->IASetInputLayout(backgroundInputLayout);
	effect->GetTechniqueByName("envmapped")->GetPassByName("background")->Apply(0, context);
	backgroundQuad->draw(context);
	#pragma endregion draw background

	Dxa::Sas::render(context);
}

void Dxa::Import::openFile()
{
	char filename[512];
	char ext[512];
	filename[0] = L'\0';
	OPENFILENAMEA of;
	of.lStructSize = sizeof(OPENFILENAME);
	of.hwndOwner = DXUTGetHWND();
	of.hInstance = NULL;
	of.lpstrFilter = "Collada (*.dae, *.xml)\0*.dae;*.xml\0Biovision BVH (*.bvh)\0*.bvh\0 3D Studio Max 3DS (*.3ds)\0*.3ds\0 3D Studio Max ASE (*.ase)\0*.ase\0Wavefront Object (*.obj)\0*.obj\0Stanford Polygon Library (*.ply)\0*.ply\0AutoCAD DXF (*.dxf)\0*.dxf\0Neutral File Format (*.nff)\0*.nff\0Sense8 WorldToolkit (*.nff)\0*.nff\0Valve Model (*.smd,*.vta)\0*.smd,*.vta\0Quake I (*.mdl)\0*.mdl\0Quake II (*.md2)\0*.md2\0Quake III (*.md3)\0*.md3\0RtCW (*.mdc)\0*.mdc\0Doom 3 (*.md5mesh;*.md5anim;*.md5camera)\0*.md5mesh;*.md5anim;*.md5camera\0DirectX X (*.x)\0*.x\0Quick3D (*.q3o, *.q3s)\0*.q3o;*q3s\0Raw Triangles (*.raw)\0*.raw\0AC3D (*.ac)\0*.ac\0Stereolithography (*.stl)\0*.stl\0Autodesk DXF (*.dxf)\0*.dxf\0Irrlicht Mesh (*.irrmesh, *.xml)\0*.irrmesh;*.xml\0Irrlicht Scene (*.irr, *.xml)\0*.irr;*.xml\0Object File Format (*.off)\0*.off\0Terragen Terrain (*.ter)\0*.ter\0 3D GameStudio Model (*.mdl)\0*.mdl\0 3D GameStudio Terrain (*.hmp)\0*.hmp\0Ogre (*.mesh.xml, *.skeleton.xml, *.material)\0*.mesh.xml;*.skeleton.xml;*.material\0Milkshape 3D (*.ms3d)\0*.ms3d\0LightWave Model (*.lwo)\0*.lwo\0LightWave Scene (*.lws)\0*.lws\0Modo Model (*.lxo)\0*.lxo\0CharacterStudio Motion (*.csm)\0*.csm\0Stanford Ply (*.ply)\0*.ply\0TrueSpace (*.cob, *.scn)\0*.cob;*.scn\0\0";
	of.lpstrCustomFilter = NULL;
	of.nMaxCustFilter = 0;
	of.nFilterIndex = 0;
	of.lpstrFile = filename;
	of.nMaxFile = 512; 
	of.lpstrFileTitle = NULL;
	of.nMaxFileTitle = 0;
	of.lpstrInitialDir = ".";
	of.lpstrTitle = NULL;
	of.Flags = 0; 
	of.nFileOffset = 0; 
	of.nFileExtension = 0; 
	of.lpstrDefExt = NULL;
	of.lCustData = NULL;
	of.lpfnHook = NULL;
	of.lpTemplateName = NULL; 
	of.pvReserved = NULL;
	of.dwReserved = NULL;
	of.FlagsEx = 0;

	GetOpenFileNameA(&of);

	if(filename[0] == '\0')
		return;

	delete mesh;
	delete vertexData;
	delete indexData;

	Assimp::Importer importer;

	const aiScene* assScene = importer.ReadFile( filename,
	//	aiProcess_FlipWindingOrder		|
		aiProcess_CalcTangentSpace		| 
		aiProcess_Triangulate			|
		aiProcess_JoinIdenticalVertices	|
		aiProcess_SortByPType);
  
	// if the import failed
	if( !assScene || !assScene->HasMeshes() || assScene->mNumMeshes == 0) 
	{
		mesh = NULL;
		MessageBoxA( NULL, 
			"Mesh could not be loaded.",
			"Failed to create mesh from file!", MB_OK);
		exit(-1);
	}

	aiMesh* assMesh = assScene->mMeshes[0];

	Mesh::importMesh(assMesh, device, &mesh, &vertexData, &indexData);
	if(mesh == NULL)
	{
		MessageBoxA( NULL, 
			"Mesh could not be loaded.",
			"Failed to create mesh from file!", MB_OK);
		exit(-1);
	}

	D3DX11_PASS_DESC basicPassDesc;
	effect->GetTechniqueByName("envmapped")->GetPassByName("envmapped")->GetDesc(&basicPassDesc);
	inputLayout = inputLayoutManager->getCompatibleInputLayout(device, basicPassDesc, mesh, false);

}

void Dxa::Import::saveFile()
{
	wchar_t filename[512];
	filename[0] = L'\0';
	OPENFILENAME of;
	of.lStructSize = sizeof(OPENFILENAME);
	of.hwndOwner = DXUTGetHWND();
	of.hInstance = NULL;
	of.lpstrFilter = L"Geometry as Buffers\0*.dgb\0\0";
	of.lpstrCustomFilter = NULL;
	of.nMaxCustFilter = 0;
	of.nFilterIndex = 0;
	of.lpstrFile = filename;
	of.nMaxFile = 512; 
	of.lpstrFileTitle = NULL;
	of.nMaxFileTitle = 0;
	of.lpstrInitialDir = L"."; 
	of.lpstrTitle = NULL;
	of.Flags = 0; 
	of.nFileOffset = 0; 
	of.nFileExtension = 0; 
	of.lpstrDefExt = L"dgb";
	of.lCustData = NULL;
	of.lpfnHook = NULL;
	of.lpTemplateName = NULL; 
	of.pvReserved = NULL;
	of.dwReserved = NULL;
	of.FlagsEx = 0;

	GetSaveFileName(&of);
	HANDLE file = CreateFile(filename, FILE_WRITE_DATA, 0, NULL, CREATE_ALWAYS, 0, NULL);
    if( INVALID_HANDLE_VALUE == file )
		return;

    // Get the file size
    LARGE_INTEGER fileSize;
    GetFileSizeEx( file, &fileSize );
    UINT cBytes = fileSize.LowPart;

	const D3D11_INPUT_ELEMENT_DESC* elements;
	unsigned int nElements;
	mesh->getElements(&elements, &nElements, false);

	for(int u=0; u<nElements; u++)
	{
		unsigned int nBytesWritten;
		WriteFile(file, (void*)&elements[u], sizeof(D3D11_INPUT_ELEMENT_DESC), (LPDWORD)&nBytesWritten, NULL);
		WriteFile(file, (void*)elements[u].SemanticName, strlen(elements[u].SemanticName) + 1, (LPDWORD)&nBytesWritten, NULL);
	}
	unsigned int nBytesWritten;
	char nullChar = 0;
	WriteFile(file, (void*)&nullChar, 1, (LPDWORD)&nBytesWritten, NULL);

	unsigned int nVertices = mesh->getVertexCount();
	unsigned int nFaces = mesh->getFaceCount();

	WriteFile(file, (void*)&nVertices, sizeof(unsigned int), (LPDWORD)&nBytesWritten, NULL);
	WriteFile(file, (void*)&nFaces, sizeof(unsigned int), (LPDWORD)&nBytesWritten, NULL);

	unsigned int vertexBufferByteLength = mesh->getVertexStride() * mesh->getVertexCount();
	WriteFile(file, vertexData, vertexBufferByteLength, (LPDWORD)&nBytesWritten, NULL);

	unsigned int indexBufferByteLength = mesh->getFaceCount() * mesh->getIndexStride() * 3;
	WriteFile(file, indexData, indexBufferByteLength, (LPDWORD)&nBytesWritten, NULL);

    CloseHandle( file );
}