#include "DXUT.h"
#include "Mass.h"
#include <assimp.hpp>      // C++ importer interface
#include <aiScene.h>       // Output data structure
#include <aiPostProcess.h> // Post processing flags


HRESULT Mass::createMassFromFile(const char* assetFilename, ID3D11Device* device, Mass** mass)
{
	Assimp::Importer importer;

	const aiScene* assScene = importer.ReadFile( assetFilename, 
        aiProcess_CalcTangentSpace       | 
        aiProcess_Triangulate            |
        aiProcess_JoinIdenticalVertices  |
        aiProcess_SortByPType);
  
	// if the import failed
	if( !assScene || !assScene->HasMeshes() || assScene->mNumMeshes == 0) 
	{
		*mass = NULL;
		return E_FAIL;
	}

	aiMesh* assMesh = assScene->mMeshes[0];

	*mass = new Mass();
	(*mass)->create(device, assMesh);

    return S_OK;
}

HRESULT  Mass::create(
		ID3D11Device* device,
		aiMesh* assMesh)
{
	D3D11_INPUT_ELEMENT_DESC elements[64];
	unsigned int cElements = 0;
	unsigned int cOffset = 0;
	unsigned int positionOffset = 0;
	unsigned int normalOffset = 0;
	unsigned int texcoord0Offset = 0;

	if(assMesh->HasPositions())
	{
		elements[cElements].Format = DXGI_FORMAT_R32G32B32_FLOAT;
		elements[cElements].AlignedByteOffset = positionOffset = cOffset;
		cOffset += sizeof(float) * 3;
		elements[cElements].InputSlot = 0;
		elements[cElements].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
		elements[cElements].InstanceDataStepRate = 0;
		elements[cElements].SemanticIndex = 0;
		const char* semanticNameLiteral = "POSITION";
		char* semanticName = new char(strlen(semanticNameLiteral)+1);
		strcpy(semanticName, semanticNameLiteral);
		elements[cElements].SemanticName = semanticName;
		cElements++;
	}
	if(assMesh->HasNormals())
	{
		elements[cElements].Format = DXGI_FORMAT_R32G32B32_FLOAT;
		elements[cElements].AlignedByteOffset = normalOffset = cOffset;
		cOffset += sizeof(float) * 3;
		elements[cElements].InputSlot = 0;
		elements[cElements].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
		elements[cElements].InstanceDataStepRate = 0;
		elements[cElements].SemanticIndex = 0;
		const char* semanticNameLiteral = "NORMAL";
		char* semanticName = new char(strlen(semanticNameLiteral)+1);
		strcpy(semanticName, semanticNameLiteral);
		elements[cElements].SemanticName = semanticName;
		cElements++;
	}
	if(assMesh->HasTextureCoords(0))
	{
		elements[cElements].Format = DXGI_FORMAT_R32G32_FLOAT;
		elements[cElements].AlignedByteOffset = texcoord0Offset = cOffset;
		cOffset += sizeof(float) * 2;
		elements[cElements].InputSlot = 0;
		elements[cElements].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
		elements[cElements].InstanceDataStepRate = 0;
		elements[cElements].SemanticIndex = 0;
		const char* semanticNameLiteral = "TEXCOORD";
		char* semanticName = new char(strlen(semanticNameLiteral)+1);
		strcpy(semanticName, semanticNameLiteral);
		elements[cElements].SemanticName = semanticName;
		cElements++;
	}
	this->vertexStride = cOffset;
	this->nElementsWithoutInstancing = cElements;

	{
		elements[cElements].Format = DXGI_FORMAT_R32G32B32A32_FLOAT;
		elements[cElements].AlignedByteOffset = 0;
		elements[cElements].InputSlot = 0;
		elements[cElements].InputSlotClass = D3D11_INPUT_PER_INSTANCE_DATA;
		elements[cElements].InstanceDataStepRate = 1;
		elements[cElements].SemanticIndex = 0;
		const char* semanticNameLiteral = "INSTANCEMODELDQ";
		char* semanticName = new char(strlen(semanticNameLiteral)+1);
		strcpy(semanticName, semanticNameLiteral);
		elements[cElements].SemanticName = semanticName;
		cElements++;
	}
	{
		elements[cElements].Format = DXGI_FORMAT_R32G32B32A32_FLOAT;
		elements[cElements].AlignedByteOffset = sizeof(float) * 4;
		elements[cElements].InputSlot = 0;
		elements[cElements].InputSlotClass = D3D11_INPUT_PER_INSTANCE_DATA;
		elements[cElements].InstanceDataStepRate = 1;
		elements[cElements].SemanticIndex = 1;
		const char* semanticNameLiteral = "INSTANCEMODELDQ";
		char* semanticName = new char(strlen(semanticNameLiteral)+1);
		strcpy(semanticName, semanticNameLiteral);
		elements[cElements].SemanticName = semanticName;
		cElements++;
	}
	{
		elements[cElements].Format = DXGI_FORMAT_R32G32B32A32_UINT;
		elements[cElements].AlignedByteOffset = sizeof(int) * 8;
		elements[cElements].InputSlot = 0;
		elements[cElements].InputSlotClass = D3D11_INPUT_PER_INSTANCE_DATA;
		elements[cElements].InstanceDataStepRate = 1;
		elements[cElements].SemanticIndex = 0;
		const char* semanticNameLiteral = "INSTANCEOFFSET";
		char* semanticName = new char(strlen(semanticNameLiteral)+1);
		strcpy(semanticName, semanticNameLiteral);
		elements[cElements].SemanticName = semanticName;
		cElements++;
	}

	this->nElements = cElements;
	this->elements = new D3D11_INPUT_ELEMENT_DESC[nElements];
	memcpy(this->elements, elements, nElements * sizeof(D3D11_INPUT_ELEMENT_DESC));

	this->nVertices = assMesh->mNumVertices;
	this->nFaces = assMesh->mNumFaces;
	this->wideIndexBuffer = nVertices > USHRT_MAX;

	D3D11_BUFFER_DESC vertexBufferDesc;
	vertexBufferDesc.BindFlags = D3D11_BIND_VERTEX_BUFFER;
	vertexBufferDesc.ByteWidth = nVertices * vertexStride;
	vertexBufferDesc.CPUAccessFlags = 0;
	vertexBufferDesc.MiscFlags = 0;
	vertexBufferDesc.StructureByteStride = vertexStride;
	vertexBufferDesc.Usage = D3D11_USAGE_IMMUTABLE;

	char* sysMemVertices = new char[nVertices * vertexStride];

	for(unsigned int iVertex=0; iVertex < assMesh->mNumVertices; iVertex++)
	{
		memcpy(sysMemVertices + iVertex * vertexStride + positionOffset, &assMesh->mVertices[iVertex], sizeof(float) * 3);
		memcpy(sysMemVertices + iVertex * vertexStride + normalOffset, &assMesh->mNormals[iVertex], sizeof(float) * 3);
		memcpy(sysMemVertices + iVertex * vertexStride + texcoord0Offset, &assMesh->mTextureCoords[0][iVertex], sizeof(float) * 2);
	}

	D3D11_SUBRESOURCE_DATA initialVertexData;
	initialVertexData.pSysMem = sysMemVertices;
	HRESULT hr = device->CreateBuffer(&vertexBufferDesc, &initialVertexData, &vertexBuffer);
	delete sysMemVertices;
	if(hr != S_OK)
		return hr;

	D3D11_BUFFER_DESC indexBufferDesc;
	indexBufferDesc.BindFlags = D3D11_BIND_INDEX_BUFFER;
	indexBufferDesc.ByteWidth = nFaces * 3 * (wideIndexBuffer?4:2);
	indexBufferDesc.CPUAccessFlags = 0;
	indexBufferDesc.MiscFlags = 0;
	indexBufferDesc.StructureByteStride = 0;;
	indexBufferDesc.Usage = D3D11_USAGE_IMMUTABLE;

	D3D11_SUBRESOURCE_DATA initialIndexData;
	if(wideIndexBuffer)
	{
		unsigned int* sysMemIndices = new unsigned int[nFaces*3];
		for(int iFace=0; iFace < nFaces; iFace++)
		{
			sysMemIndices[iFace * 3 + 0] = assMesh->mFaces[iFace].mIndices[0];
			sysMemIndices[iFace * 3 + 1] = assMesh->mFaces[iFace].mIndices[1];
			sysMemIndices[iFace * 3 + 2] = assMesh->mFaces[iFace].mIndices[2];
		}
		initialIndexData.pSysMem = sysMemIndices;
	}
	else
	{
		unsigned short* sysMemIndices = new unsigned short[nFaces*3];
		for(int iFace=0; iFace < nFaces; iFace++)
		{
			sysMemIndices[iFace * 3 + 0] = assMesh->mFaces[iFace].mIndices[0];
			sysMemIndices[iFace * 3 + 1] = assMesh->mFaces[iFace].mIndices[1];
			sysMemIndices[iFace * 3 + 2] = assMesh->mFaces[iFace].mIndices[2];
		}
		initialIndexData.pSysMem = sysMemIndices;
	}
	
	hr = device->CreateBuffer(&indexBufferDesc, &initialIndexData, &indexBuffer);
	delete initialIndexData.pSysMem;
	if(hr != S_OK)
		return hr;

	this->topology = D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST;
	
	assMesh->

	return hr;
}

Mass::~Mass(void)
{
	for(int i=0; i<nElements; i++)
		if(elements[i].SemanticName)
			delete elements[i].SemanticName;
	delete [] elements;
	if(indexBuffer)
		indexBuffer->Release();
	if(vertexBuffer)
		vertexBuffer->Release();
}

void Mass::getElementsWithoutInstancing(const D3D11_INPUT_ELEMENT_DESC** elements, unsigned int* nElements)
{
	*elements = this->elements;
	*nElements = this->nElementsWithoutInstancing;
}

void Mass::getElements(const D3D11_INPUT_ELEMENT_DESC** elements, unsigned int* nElements)
{
	*elements = this->elements;
	*nElements = this->nElements;
}