#include "DXUT.h"
#include "Mesh.h"
#include <assimp.hpp>      // C++ importer interface
#include <aiScene.h>       // Output data structure
#include <aiPostProcess.h> // Post processing flags

HRESULT Mesh::createMeshFromFile(const char* dgbFileName, ID3D11Device* device, Mesh** mesh)
{
    // Open the file
    HANDLE file = CreateFileA(dgbFileName, FILE_READ_DATA, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
    if( INVALID_HANDLE_VALUE == file )
        return DXUTERR_MEDIANOTFOUND;

    // Get the file size
    LARGE_INTEGER fileSize;
    GetFileSizeEx( file, &fileSize );
    UINT cBytes = fileSize.LowPart;

    // Allocate memory
    BYTE* meshData = new BYTE[ cBytes ];
    if( !meshData )
    {
        CloseHandle( file );
        return E_OUTOFMEMORY;
    }

    // Read in the file
	HRESULT hr = S_OK;
    DWORD dwBytesRead;
    if( !ReadFile( file, meshData, cBytes, &dwBytesRead, NULL ) )
        hr = E_FAIL;

    CloseHandle( file );

	if( SUCCEEDED(hr) )
    {
        hr = createMeshFromMemory(  device,
                                meshData,
                                cBytes,
								mesh);
        if( FAILED(hr) )
            delete [] meshData;
    }
	delete [] meshData;
    return hr;
}

HRESULT Mesh::createMeshFromMemory(ID3D11Device* device, BYTE* data, unsigned int nBytes, Mesh** mesh)
{
	D3D11_INPUT_ELEMENT_DESC elements[20];
	
	BYTE* p = data;

	unsigned int nElements = 0;
	while(*p)
	{
		CopyMemory((void*)&elements[nElements], (void*)p, sizeof(D3D11_INPUT_ELEMENT_DESC));
		p += sizeof(D3D11_INPUT_ELEMENT_DESC);
		unsigned int semanticNameLength = strlen( (char*)p) + 1;
		char* semanticName = new char[semanticNameLength];
		strcpy(semanticName, (char*)p);
		p += semanticNameLength;
		elements[nElements].SemanticName = semanticName;
		nElements++;
	}
	p++;

	unsigned int nVertices = *(unsigned int*)p;
	p += sizeof(unsigned int);
	unsigned int nFaces = *(unsigned int*)p;
	p += sizeof(unsigned int);

	bool wideIndexBuffer = false;
	if(nFaces * 3 >= 0xffff)
		wideIndexBuffer = true;

	if(*mesh == NULL)
		*mesh =	new Mesh();
	(*mesh)->create(device, elements, nElements, nVertices, p, nFaces, NULL, wideIndexBuffer);


	for(int iElements=0; iElements<nElements; iElements++)
		delete elements[iElements].SemanticName;
	return S_OK;
}

HRESULT Mesh::createQuad(ID3D11Device* device, Mesh** mesh)
{
	D3D11_INPUT_ELEMENT_DESC quadElements[] =
	{
		{ "POSITION", 0, DXGI_FORMAT_R32G32B32A32_FLOAT, 0, 0, D3D11_INPUT_PER_VERTEX_DATA, 0 },
		{ "TEXCOORD", 0, DXGI_FORMAT_R32G32_FLOAT, 0, 16, D3D11_INPUT_PER_VERTEX_DATA, 0 }
	};

	struct QuadVertex
	{
		D3DXVECTOR4 pos;
		D3DXVECTOR2 tex;
	} svQuad[4];
	static const float fSize = 1.0f;
	svQuad[0].pos = D3DXVECTOR4(-fSize, fSize, 0.9999f, 1.0f);
	svQuad[0].tex = D3DXVECTOR2(0.0f, 0.0f);
	svQuad[1].pos = D3DXVECTOR4(fSize, fSize, 0.9999f, 1.0f);
	svQuad[1].tex = D3DXVECTOR2(1.0f, 0.0f);
	svQuad[2].pos = D3DXVECTOR4(-fSize, -fSize, 0.9999f, 1.0f);
	svQuad[2].tex = D3DXVECTOR2(0.0f, 1.0f);
	svQuad[3].pos = D3DXVECTOR4(fSize, -fSize, 0.9999f, 1.0f);
	svQuad[3].tex = D3DXVECTOR2(1.0f, 1.0f);

	unsigned short siQuad[6];
	siQuad[0] = 0;
	siQuad[1] = 1;
	siQuad[2] = 2;
	siQuad[3] = 3;
	siQuad[4] = 2;
	siQuad[5] = 1;

	*mesh = new Mesh();
	
	return (*mesh)->create(device, quadElements, 2, 4, svQuad, 2, siQuad, false);
}

Mesh::Mesh(void)
{
	vertexBuffer = NULL;
	indexBuffer = NULL;
	elements = NULL;
	nElements = 0;

	vertexStride = 0;
	wideIndexBuffer = false;

	nVertices = 0;
	nFaces = 0;

	topology = D3D11_PRIMITIVE_TOPOLOGY_UNDEFINED;
}

HRESULT Mesh::create(
		ID3D11Device* device,
		D3D11_INPUT_ELEMENT_DESC* elements,
		unsigned int nElements,
		unsigned int nVertices,
		void* vertexData,
		unsigned int nFaces,
		void* indexData,
		bool wideIndexBuffer)
{
	HRESULT hr = S_OK;

	this->nElements = nElements;
	this->elements = new D3D11_INPUT_ELEMENT_DESC[nElements + 3];
	memcpy(this->elements, elements, nElements * sizeof(D3D11_INPUT_ELEMENT_DESC));

	this->nVertices = nVertices;
	this->nFaces = nFaces;
	this->wideIndexBuffer = wideIndexBuffer;

	vertexStride = 0;
	for(unsigned int iElement=0; iElement<nElements; iElement++)
	{
		char* semanticnameCopy = new char[strlen(elements[iElement].SemanticName)+1];
		strcpy(semanticnameCopy, elements[iElement].SemanticName);
		this->elements[iElement].SemanticName = semanticnameCopy;
		switch(elements[iElement].Format)
		{
			case DXGI_FORMAT_R32_SINT :
			case DXGI_FORMAT_R32_UINT :
			case DXGI_FORMAT_R32_FLOAT : vertexStride += sizeof(float); break;
			case DXGI_FORMAT_R32G32_SINT:
			case DXGI_FORMAT_R32G32_UINT:
			case DXGI_FORMAT_R32G32_FLOAT : vertexStride += 2 * sizeof(float); break;
			case DXGI_FORMAT_R32G32B32_SINT:
			case DXGI_FORMAT_R32G32B32_UINT:
			case DXGI_FORMAT_R32G32B32_FLOAT : vertexStride += 3 * sizeof(float); break;
			case DXGI_FORMAT_R32G32B32A32_SINT:
			case DXGI_FORMAT_R32G32B32A32_UINT:
			case DXGI_FORMAT_R32G32B32A32_FLOAT : vertexStride += 4 * sizeof(float); break;
			default:
				bool unknown = true; break;
		};
	}

	unsigned int cElements = nElements;
	{
		this->elements[cElements].Format = DXGI_FORMAT_R32G32B32A32_FLOAT;
		this->elements[cElements].AlignedByteOffset = 0;
		this->elements[cElements].InputSlot = 1;
		this->elements[cElements].InputSlotClass = D3D11_INPUT_PER_INSTANCE_DATA;
		this->elements[cElements].InstanceDataStepRate = 1;
		this->elements[cElements].SemanticIndex = 0;
		const char* semanticNameLiteral = "INSTANCEMODELDQ";
		char* semanticName = new char[strlen(semanticNameLiteral)+1];
		strcpy(semanticName, semanticNameLiteral);
		this->elements[cElements].SemanticName = semanticName;
		cElements++;
	}
	{
		this->elements[cElements].Format = DXGI_FORMAT_R32G32B32A32_FLOAT;
		this->elements[cElements].AlignedByteOffset = sizeof(float) * 4;
		this->elements[cElements].InputSlot = 1;
		this->elements[cElements].InputSlotClass = D3D11_INPUT_PER_INSTANCE_DATA;
		this->elements[cElements].InstanceDataStepRate = 1;
		this->elements[cElements].SemanticIndex = 1;
		const char* semanticNameLiteral = "INSTANCEMODELDQ";
		char* semanticName = new char[strlen(semanticNameLiteral)+1];
		strcpy(semanticName, semanticNameLiteral);
		this->elements[cElements].SemanticName = semanticName;
		cElements++;
	}
	{
		this->elements[cElements].Format = DXGI_FORMAT_R32G32B32A32_UINT;
		this->elements[cElements].AlignedByteOffset = sizeof(int) * 8;
		this->elements[cElements].InputSlot = 1;
		this->elements[cElements].InputSlotClass = D3D11_INPUT_PER_INSTANCE_DATA;
		this->elements[cElements].InstanceDataStepRate = 1;
		this->elements[cElements].SemanticIndex = 0;
		const char* semanticNameLiteral = "INSTANCEOFFSET";
		char* semanticName = new char[strlen(semanticNameLiteral)+1];
		strcpy(semanticName, semanticNameLiteral);
		this->elements[cElements].SemanticName = semanticName;
		cElements++;
	}
	this->nElementsInstancing = cElements;

	D3D11_BUFFER_DESC vertexBufferDesc;
	vertexBufferDesc.BindFlags = D3D11_BIND_VERTEX_BUFFER;
	vertexBufferDesc.ByteWidth = nVertices * vertexStride;
	vertexBufferDesc.CPUAccessFlags = 0;
	vertexBufferDesc.MiscFlags = 0;
	vertexBufferDesc.StructureByteStride = vertexStride;
	vertexBufferDesc.Usage = D3D11_USAGE_IMMUTABLE;

	D3D11_SUBRESOURCE_DATA initialVertexData;
	initialVertexData.pSysMem = vertexData;

	hr = device->CreateBuffer(&vertexBufferDesc, &initialVertexData, &vertexBuffer);
	if(hr != S_OK)
		return hr;

	D3D11_BUFFER_DESC indexBufferDesc;
	indexBufferDesc.BindFlags = D3D11_BIND_INDEX_BUFFER;
	indexBufferDesc.ByteWidth = nFaces * 3 * (wideIndexBuffer?4:2);
	indexBufferDesc.CPUAccessFlags = 0;
	indexBufferDesc.MiscFlags = 0;
	indexBufferDesc.StructureByteStride = 0;//wideIndexBuffer?4:2;
	indexBufferDesc.Usage = D3D11_USAGE_IMMUTABLE;

	D3D11_SUBRESOURCE_DATA initialIndexData;
	initialIndexData.pSysMem = indexData;
	if(initialIndexData.pSysMem == NULL)
		initialIndexData.pSysMem = (char*)vertexData + vertexBufferDesc.ByteWidth;
	
	hr = device->CreateBuffer(&indexBufferDesc, &initialIndexData, &indexBuffer);

	this->topology = D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST;

	return hr;
}

HRESULT Mesh::importMesh(aiMesh* assMesh, ID3D11Device* device, Mesh** mesh, const void** vertexData, const void** indexData)
{
	*mesh = new Mesh();
	return (*mesh)->import(device, assMesh, vertexData, indexData);
}


HRESULT  Mesh::import(
		ID3D11Device* device,
		aiMesh* assMesh, const void** vertexData, const void** indexData)
{
	D3D11_INPUT_ELEMENT_DESC elements[64];
	unsigned int cElements = 0;
	unsigned int cOffset = 0;
	unsigned int positionOffset = 0;
	unsigned int normalOffset = 0;
	unsigned int tangentOffset = 0;
	unsigned int binormalOffset = 0;
	unsigned int texcoord0Offset = 0;

	if(assMesh->HasPositions())
	{
		elements[cElements].Format = DXGI_FORMAT_R32G32B32_FLOAT;
		elements[cElements].AlignedByteOffset = positionOffset = cOffset;
		cOffset += sizeof(float) * 3;
		elements[cElements].InputSlot = 0;
		elements[cElements].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
		elements[cElements].InstanceDataStepRate = 0;
		elements[cElements].SemanticIndex = 0;
		const char* semanticNameLiteral = "POSITION";
		char* semanticName = new char[strlen(semanticNameLiteral)+1];
		strcpy(semanticName, semanticNameLiteral);
		elements[cElements].SemanticName = semanticName;
		cElements++;
	}
	if(assMesh->HasNormals())
	{
		elements[cElements].Format = DXGI_FORMAT_R32G32B32_FLOAT;
		elements[cElements].AlignedByteOffset = normalOffset = cOffset;
		cOffset += sizeof(float) * 3;
		elements[cElements].InputSlot = 0;
		elements[cElements].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
		elements[cElements].InstanceDataStepRate = 0;
		elements[cElements].SemanticIndex = 0;
		const char* semanticNameLiteral = "NORMAL";
		char* semanticName = new char[strlen(semanticNameLiteral)+1];
		strcpy(semanticName, semanticNameLiteral);
		elements[cElements].SemanticName = semanticName;
		cElements++;
	}
	if(assMesh->HasTangentsAndBitangents())
	{
		elements[cElements].Format = DXGI_FORMAT_R32G32B32_FLOAT;
		elements[cElements].AlignedByteOffset = tangentOffset = cOffset;
		cOffset += sizeof(float) * 3;
		elements[cElements].InputSlot = 0;
		elements[cElements].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
		elements[cElements].InstanceDataStepRate = 0;
		elements[cElements].SemanticIndex = 0;
		const char* semanticNameLiteral = "TANGENT";
		char* semanticName = new char[strlen(semanticNameLiteral)+1];
		strcpy(semanticName, semanticNameLiteral);
		elements[cElements].SemanticName = semanticName;
		cElements++;
	}
	if(assMesh->HasTangentsAndBitangents())
	{
		elements[cElements].Format = DXGI_FORMAT_R32G32B32_FLOAT;
		elements[cElements].AlignedByteOffset = binormalOffset = cOffset;
		cOffset += sizeof(float) * 3;
		elements[cElements].InputSlot = 0;
		elements[cElements].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
		elements[cElements].InstanceDataStepRate = 0;
		elements[cElements].SemanticIndex = 0;
		const char* semanticNameLiteral = "BINORMAL";
		char* semanticName = new char[strlen(semanticNameLiteral)+1];
		strcpy(semanticName, semanticNameLiteral);
		elements[cElements].SemanticName = semanticName;
		cElements++;
	}
	if(assMesh->HasTextureCoords(0))
	{
		elements[cElements].Format = DXGI_FORMAT_R32G32_FLOAT;
		elements[cElements].AlignedByteOffset = texcoord0Offset = cOffset;
		cOffset += sizeof(float) * 2;
		elements[cElements].InputSlot = 0;
		elements[cElements].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
		elements[cElements].InstanceDataStepRate = 0;
		elements[cElements].SemanticIndex = 0;
		const char* semanticNameLiteral = "TEXCOORD";
		char* semanticName = new char[strlen(semanticNameLiteral)+1];
		strcpy(semanticName, semanticNameLiteral);
		elements[cElements].SemanticName = semanticName;
		cElements++;
	}
	this->vertexStride = cOffset;
	this->nElements = cElements;

	{
		elements[cElements].Format = DXGI_FORMAT_R32G32B32A32_FLOAT;
		elements[cElements].AlignedByteOffset = 0;
		elements[cElements].InputSlot = 1;
		elements[cElements].InputSlotClass = D3D11_INPUT_PER_INSTANCE_DATA;
		elements[cElements].InstanceDataStepRate = 1;
		elements[cElements].SemanticIndex = 0;
		const char* semanticNameLiteral = "INSTANCEMODELDQ";
		char* semanticName = new char[strlen(semanticNameLiteral)+1];
		strcpy(semanticName, semanticNameLiteral);
		elements[cElements].SemanticName = semanticName;
		cElements++;
	}
	{
		elements[cElements].Format = DXGI_FORMAT_R32G32B32A32_FLOAT;
		elements[cElements].AlignedByteOffset = sizeof(float) * 4;
		elements[cElements].InputSlot = 1;
		elements[cElements].InputSlotClass = D3D11_INPUT_PER_INSTANCE_DATA;
		elements[cElements].InstanceDataStepRate = 1;
		elements[cElements].SemanticIndex = 1;
		const char* semanticNameLiteral = "INSTANCEMODELDQ";
		char* semanticName = new char[strlen(semanticNameLiteral)+1];
		strcpy(semanticName, semanticNameLiteral);
		elements[cElements].SemanticName = semanticName;
		cElements++;
	}
	{
		elements[cElements].Format = DXGI_FORMAT_R32G32B32A32_UINT;
		elements[cElements].AlignedByteOffset = sizeof(int) * 8;
		elements[cElements].InputSlot = 1;
		elements[cElements].InputSlotClass = D3D11_INPUT_PER_INSTANCE_DATA;
		elements[cElements].InstanceDataStepRate = 1;
		elements[cElements].SemanticIndex = 0;
		const char* semanticNameLiteral = "INSTANCEOFFSET";
		char* semanticName = new char[strlen(semanticNameLiteral)+1];
		strcpy(semanticName, semanticNameLiteral);
		elements[cElements].SemanticName = semanticName;
		cElements++;
	}
	this->nElementsInstancing = cElements;

	this->elements = new D3D11_INPUT_ELEMENT_DESC[nElementsInstancing];
	memcpy(this->elements, elements, nElementsInstancing * sizeof(D3D11_INPUT_ELEMENT_DESC));

	this->nVertices = assMesh->mNumVertices;
	this->nFaces = assMesh->mNumFaces;
	this->wideIndexBuffer = nVertices > USHRT_MAX;

	D3D11_BUFFER_DESC vertexBufferDesc;
	vertexBufferDesc.BindFlags = D3D11_BIND_VERTEX_BUFFER;
	vertexBufferDesc.ByteWidth = nVertices * vertexStride;
	vertexBufferDesc.CPUAccessFlags = 0;
	vertexBufferDesc.MiscFlags = 0;
	vertexBufferDesc.StructureByteStride = vertexStride;
	vertexBufferDesc.Usage = D3D11_USAGE_IMMUTABLE;

	char* sysMemVertices = new char[nVertices * vertexStride];

	for(unsigned int iVertex=0; iVertex < assMesh->mNumVertices; iVertex++)
	{
		memcpy(sysMemVertices + iVertex * vertexStride + positionOffset, &assMesh->mVertices[iVertex], sizeof(float) * 3);
		if(assMesh->HasNormals())
			memcpy(sysMemVertices + iVertex * vertexStride + normalOffset, &assMesh->mNormals[iVertex], sizeof(float) * 3);
		if(assMesh->HasTangentsAndBitangents())
		{
			memcpy(sysMemVertices + iVertex * vertexStride + tangentOffset, &assMesh->mTangents[iVertex], sizeof(float) * 3);
			memcpy(sysMemVertices + iVertex * vertexStride + binormalOffset, &assMesh->mBitangents[iVertex], sizeof(float) * 3);
		}
		if(assMesh->HasTextureCoords(0))
			memcpy(sysMemVertices + iVertex * vertexStride + texcoord0Offset, &assMesh->mTextureCoords[0][iVertex], sizeof(float) * 2);
	}

	D3D11_SUBRESOURCE_DATA initialVertexData;
	initialVertexData.pSysMem = sysMemVertices;
	HRESULT hr = device->CreateBuffer(&vertexBufferDesc, &initialVertexData, &vertexBuffer);
	if(vertexData)
		*vertexData = sysMemVertices;
	else
		delete sysMemVertices;
	if(hr != S_OK)
		return hr;

	D3D11_BUFFER_DESC indexBufferDesc;
	indexBufferDesc.BindFlags = D3D11_BIND_INDEX_BUFFER;
	indexBufferDesc.ByteWidth = nFaces * 3 * (wideIndexBuffer?4:2);
	indexBufferDesc.CPUAccessFlags = 0;
	indexBufferDesc.MiscFlags = 0;
	indexBufferDesc.StructureByteStride = 0;;
	indexBufferDesc.Usage = D3D11_USAGE_IMMUTABLE;

	D3D11_SUBRESOURCE_DATA initialIndexData;
	if(wideIndexBuffer)
	{
		unsigned int* sysMemIndices = new unsigned int[nFaces*3];
		for(int iFace=0; iFace < nFaces; iFace++)
		{
			sysMemIndices[iFace * 3 + 0] = assMesh->mFaces[iFace].mIndices[0];
			sysMemIndices[iFace * 3 + 1] = assMesh->mFaces[iFace].mIndices[1];
			sysMemIndices[iFace * 3 + 2] = assMesh->mFaces[iFace].mIndices[2];
		}
		initialIndexData.pSysMem = sysMemIndices;
	}
	else
	{
		unsigned short* sysMemIndices = new unsigned short[nFaces*3];
		for(int iFace=0; iFace < nFaces; iFace++)
		{
			sysMemIndices[iFace * 3 + 0] = assMesh->mFaces[iFace].mIndices[0];
			sysMemIndices[iFace * 3 + 1] = assMesh->mFaces[iFace].mIndices[1];
			sysMemIndices[iFace * 3 + 2] = assMesh->mFaces[iFace].mIndices[2];
		}
		initialIndexData.pSysMem = sysMemIndices;
	}
	
	hr = device->CreateBuffer(&indexBufferDesc, &initialIndexData, &indexBuffer);
	if(indexData)
		*indexData = initialIndexData.pSysMem;
	else
		delete initialIndexData.pSysMem;
	if(hr != S_OK)
		return hr;

	this->topology = D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST;
	
	return hr;
}

Mesh::~Mesh(void)
{
	for(int i=0; i<nElementsInstancing; i++)
		if(elements[i].SemanticName)
			delete elements[i].SemanticName;
	delete [] elements;
	if(indexBuffer)
		indexBuffer->Release();
	vertexBuffer->Release();
}

HRESULT Mesh::createInputLayout(ID3D11Device* device, D3DX11_PASS_DESC* passDesc, ID3D11InputLayout** inputLayout)
{
	return device->CreateInputLayout(
		elements,
		nElements,
		passDesc->pIAInputSignature,
		passDesc->IAInputSignatureSize,
		inputLayout);
}

void Mesh::getElements(const D3D11_INPUT_ELEMENT_DESC** elements, unsigned int* nElements, bool instancing)
{
	*elements = this->elements;
	*nElements = instancing?(this->nElementsInstancing):(this->nElements);
}

void Mesh::draw(ID3D11DeviceContext* context)
{
	unsigned int offset = 0;
	context->IASetVertexBuffers(0, 1, &vertexBuffer, &vertexStride,	&offset);
	if(indexBuffer)
		context->IASetIndexBuffer(indexBuffer, wideIndexBuffer?DXGI_FORMAT_R32_UINT:DXGI_FORMAT_R16_UINT, 0);
	context->IASetPrimitiveTopology(topology);
	if(indexBuffer)
		context->DrawIndexed(nFaces * 3, 0, 0);
	else
		context->Draw(nVertices, 0);
}