#include "envmapped.fx"

struct IaosTangent
{
    float4 pos			: POSITION;
    float3 normal		: NORMAL;
    float3 tangent		: TANGENT;
    float3 binormal		: BINORMAL;
    float2 tex			: TEXCOORD;
};

struct VsosTangent
{
    float4 pos			: SV_POSITION;
	float4 worldPos		: WORLDPOS;
    float3 normal		: NORMAL;
    float3 tangent		: TANGENT;
    float3 binormal		: BINORMAL;
    float2 tex			: TEXCOORD;
};

VsosTangent vsTangent(IaosTangent input)
{
	VsosTangent output = (VsosTangent)0;
	output.pos = mul(input.pos,
		modelViewProjMatrix);
	output.worldPos = mul(input.pos,
		modelMatrix);
	output.normal = mul(modelMatrixInverse,
		float4(input.normal.xyz, 0.0));
	output.tangent = mul(modelMatrixInverse,
		float4(input.tangent.xyz, 0.0));
	output.binormal = mul(modelMatrixInverse,
		float4(input.binormal.xyz, 0.0));
	output.tex = input.tex;
	return output;
}

Texture2D normalTexture;
Texture2D bumpTexture;

float normalStrength
<
	string SasUiControl = "Slider";
	string SasUiLabel = "normal map strength";
	float SasUiMin = -1;
	float SasUiMax = 1;
	int SasUiSteps = 100;
>  = 1;

float bumpHeight
<
	string SasUiControl = "Slider";
	string SasUiLabel = "bump map strength";
	float SasUiMin = 0;
	float SasUiMax = 1;
	int SasUiSteps = 100;
>  = 0.1;

float textureScale
<
	string SasUiControl = "Slider";
	string SasUiLabel = "texture scale";
	float SasUiMin = 1;
	float SasUiMax = 100;
	int SasUiSteps = 100;
>  = 10;

float envMapWeight
<
	string SasUiControl = "Slider";
	string SasUiLabel = "reflection weight";
	float SasUiMin = 0;
	float SasUiMax = 1;
	int SasUiSteps = 100;
>  = 0.5;

float4 psNormalMapped(VsosTangent input) : SV_Target
{
	input.tex *= textureScale;

	float3x3 tbnFrame = float3x3( normalize(input.tangent), normalize(input.binormal), normalize(input.normal));

	// perturbed normal in tangent space
	float3 tNormal = (normalTexture.Sample(linearSampler, input.tex) - float3(0.5, 0.5, 0)) * float3(-normalStrength, -normalStrength, 1);

	// perturbed normal in world space
	float3 normal = mul(normalize(tNormal), tbnFrame);

	float3 viewDir = normalize(eyePos - input.worldPos);
	float3 reflectionDir = reflect(-viewDir, normal);
	return 
		envMapWeight * envTexture.Sample(linearSampler, reflectionDir) +
		(1 - envMapWeight) * kdTexture.Sample(linearSampler, input.tex) * saturate(normal.y);
}

float4 psParallax(VsosTangent input) : SV_Target
{
	input.tex *= textureScale;

	float3x3 tbnFrame = float3x3( normalize(input.tangent), normalize(input.binormal), normalize(input.normal));

	// view dir in world space
	float3 viewDir = normalize(eyePos - input.worldPos);

	// view dir in tangent space (the same as texture space + height)
	float3 tViewDir = mul(tbnFrame, viewDir);

	float bump = bumpTexture.Sample(linearSampler, input.tex);

	// texcoord modification
	float2 offset = tViewDir.xy  /*/ tViewDir.z*/ * bumpHeight;
	input.tex += offset;

	// perturbed normal in tangent space
	float3 tNormal = (normalTexture.Sample(linearSampler, input.tex) - float3(0.5, 0.5, 0)) * float3(-normalStrength, -normalStrength, 1);

	// perturbed normal in world space
	float3 normal = mul(normalize(tNormal), tbnFrame);

	float3 reflectionDir = reflect(-viewDir, normal);
	return 
		envMapWeight * envTexture.Sample(linearSampler, reflectionDir) +
		(1 - envMapWeight) * kdTexture.Sample(linearSampler, input.tex) * saturate(normal.y);

}

float4 psBinaryRelief(VsosTangent input) : SV_Target
{
	input.tex *= textureScale;

	float3x3 tbnFrame = float3x3( normalize(input.tangent), normalize(input.binormal), normalize(input.normal));

	// view dir in world space
	float3 viewDir = normalize(eyePos - input.worldPos);

	// view dir in tangent space (the same as texture space + height)
	float3 tViewDir = mul(tbnFrame, viewDir);

	float3 sRange = - tViewDir * bumpHeight / tViewDir.z * 0.5;
	float3 sPos = float3(input.tex, 0) - sRange;
	
	for( int i=0; i<16; i++ )
	{
		float bump = bumpTexture.Sample(linearSampler, sPos.xy) / textureScale;
		sRange *= 0.5;
		if (sPos.z > bump * bumpHeight) // If outside
			sPos += sRange; // Move forward
		else
			sPos -= sRange; // Move backward
	}
	input.tex = sPos.xy;

	// perturbed normal in tangent space
	float3 tNormal = (normalTexture.Sample(linearSampler, input.tex) - float3(0.5, 0.5, 0)) * float3(-normalStrength, -normalStrength, 1);

	// perturbed normal in world space
	float3 normal = mul(normalize(tNormal), tbnFrame);

	float3 reflectionDir = reflect(-viewDir, normal);
	return 
		envMapWeight * envTexture.Sample(linearSampler, reflectionDir) +
		(1 - envMapWeight) * kdTexture.Sample(linearSampler, input.tex) * saturate(normal.y);

}

technique11 normalmapped
{
	pass normalmapped
	{
		SetVertexShader ( CompileShader( vs_5_0, vsTangent() ) );
//		SetPixelShader( CompileShader( ps_5_0, psNormalMapped() ) );
//		SetPixelShader( CompileShader( ps_5_0, psParallax() ) );
		SetPixelShader( CompileShader( ps_5_0, psBinaryRelief() ) );
	}
}